import logging
import time
from abc import ABCMeta
from datetime import datetime, timezone
from logging import CRITICAL, DEBUG, ERROR, INFO, Logger, WARNING

from .config import CONFIG_LOG_MESSAGE_STATUS_KEY
from .exchange.lynceus_exchange import LynceusExchange
from .lynceus import LynceusSession
from .lynceus_message_status import LynceusMessageStatus


class LynceusClientLogExchangeWrapper(Logger):
    # Defines a threshold for automatic message status definition (can be: 0, WARNING, or ERROR).
    AUTOMATIC_MESSAGE_STATUS_LOG_LEVEL_THRESHOLD: int = ERROR

    # N.B.: Logger has already been initialized, and specified as argument, so we do NOT want to call super __init__ method.
    # pylint: disable=super-init-not-called
    def __init__(self, logger: Logger, lynceus_exchange: LynceusExchange | None):
        self.__logger: Logger = logger
        self._lynceus_exchange: LynceusExchange | None = lynceus_exchange

    # Adds various properties wrapping around internal logger property allowing to use custom Level like TRACE.
    @property
    def name(self) -> str:
        return self.__logger.name

    @property
    def level(self) -> int:
        return self.__logger.level

    @property
    def parent(self):
        return self.__logger.parent

    @property
    def propagate(self) -> bool:
        return self.__logger.propagate

    @property
    def handlers(self) -> list[logging.Handler]:
        return self.__logger.handlers

    @property
    def filters(self):
        return self.__logger.filters

    @property
    def disabled(self) -> bool:
        return self.__logger.disabled

    @property
    def root(self):
        return self.__logger.root

    @property
    def manager(self) -> logging.Manager:
        return self.__logger.manager

    @property
    def _cache(self):
        # pylint: disable=protected-access
        return self.__logger._cache

    def __manage_logging(self, log_level: int, msg, *args, **kwargs):
        message_status: LynceusMessageStatus | None = kwargs.pop(CONFIG_LOG_MESSAGE_STATUS_KEY, None)

        # Defines automatically message status for warning and error.
        if not message_status \
                and log_level >= LynceusClientLogExchangeWrapper.AUTOMATIC_MESSAGE_STATUS_LOG_LEVEL_THRESHOLD:
            message_status = LynceusMessageStatus.ERROR if log_level == ERROR else LynceusMessageStatus.WARNING

        if self._lynceus_exchange:
            # Creates an **object** somehow like Log Record, which will be consumed by caller (like API).
            # queue_message: LogRecord = self.__logger.makeRecord(self.__logger.name, log_level, None, None, msg, args, exc_info=None, func=None, extra=None, sinfo=None)
            queue_message = {
                'asctime': datetime.now(tz=timezone.utc).strftime('%Y/%m/%d %H:%M:%S'),
                'created': time.time(),
                'levelname': logging.getLevelName(log_level),
                'name': self.__logger.name,
                'message': msg,
                'message_status': message_status
            }

            # Puts the message in Lynceus exchange.
            self._lynceus_exchange.put_nowait(queue_message)

        # Requests true logging anyway.
        if message_status:
            msg += f' [{CONFIG_LOG_MESSAGE_STATUS_KEY}={message_status}]'

        self.__logger.log(log_level, msg, *args, **kwargs)

    def trace(self, msg, *args, **kwargs):
        # Intercepts the call, to registers message in the lynceus exchange.
        self.__manage_logging(LynceusSession.TRACE, msg, *args, **kwargs)

    def debug(self, msg, *args, **kwargs):
        # Intercepts the call, to registers message in the lynceus exchange.
        self.__manage_logging(DEBUG, msg, *args, **kwargs)

    def info(self, msg, *args, **kwargs):
        # Intercepts the call, to registers message in the lynceus exchange.
        self.__manage_logging(INFO, msg, *args, **kwargs)

    def warning(self, msg, *args, **kwargs):
        # Intercepts the call, to registers message in the lynceus exchange.
        self.__manage_logging(WARNING, msg, *args, **kwargs)

    def error(self, msg, *args, **kwargs):
        # Intercepts the call, to registers message in the lynceus exchange.
        self.__manage_logging(ERROR, msg, *args, **kwargs)

    def critical(self, msg, *args, **kwargs):
        # Intercepts the call, to registers message in the lynceus exchange.
        self.__manage_logging(CRITICAL, msg, *args, **kwargs)


class LynceusClientClass(metaclass=ABCMeta):
    def __init__(self, lynceus_session: LynceusSession, logger_name: str, lynceus_exchange: LynceusExchange | None):
        self._lynceus_session: LynceusSession = lynceus_session
        self._lynceus_exchange = lynceus_exchange
        self.__logger: Logger = self._lynceus_session.get_logger(logger_name)

        # Declares the self._logger attribute with Lynceus wrapper (with/out lynceus_exchange).
        self._logger: LynceusClientLogExchangeWrapper = LynceusClientLogExchangeWrapper(self.__logger, self._lynceus_exchange)



import inspect
from pathlib import Path

from lynceus.core.config import (CONFIG_GENERAL_KEY,
                                 CONFIG_PROJECT_DYNAMIC_STORAGE_MANDATORY_PARAM_MAP,
                                 CONFIG_STORAGE_DYNAMIC_TYPE,
                                 CONFIG_STORAGE_IS_DYNAMIC,
                                 CONFIG_STORAGE_REMOTE_TYPE)
from lynceus.core.config.lynceus_config import LynceusConfig
from lynceus.core.exchange.lynceus_exchange import LynceusExchange
from lynceus.core.lynceus import LynceusSession
from lynceus.core.lynceus_client import LynceusClientClass
from lynceus.files.lynceus_file import (LynceusFile,
                                        _LocalLynceusFile,
                                        _RemoteS3LynceusFile)
from lynceus.files.remote.s3 import S3Utils
from lynceus.files.storage_metadata import (ActivityStorageMetadata,
                                            GroupStorageMetadata,
                                            FileStorageMetadata,
                                            OrganizationStorageMetadata,
                                            ResourceConsumptionStorageMetadata,
                                            StorageMetadataBase,
                                            UserStorageMetadata)
from lynceus.lynceus_exceptions import LynceusFileError
from lynceus.utils import lookup_root_path


# pylint: disable=too-many-instance-attributes
class LynceusFileFactory(LynceusClientClass):
    """
    LynceusFileFactory is useful to create instance of Local or Remote according to various configuration provided once
    for all in Factory constructor.
    By default, it is configured to read and write parquet files, but it can be configured to manage any kind of file.
    """

    REMOTE_STORAGE_TYPE_S3: str = 's3'
    REMOTE_STORAGE_TYPE_SUPPORTED_LIST: set[str] = {REMOTE_STORAGE_TYPE_S3}

    # Defines supported dynamic remote storage.
    REMOTE_DYNAMIC_TYPE_FILE: str = 'files'
    REMOTE_DYNAMIC_TYPE_ORGANIZATION: str = 'organization'
    REMOTE_DYNAMIC_TYPE_ACTIVITY: str = 'activity'
    REMOTE_DYNAMIC_TYPE_USER: str = 'user'
    REMOTE_DYNAMIC_TYPE_GROUP: str = 'group'
    REMOTE_DYNAMIC_TYPE_RESOURCES_STATS: str = 'resources_stats'

    # Defines once for all classes, and awaited parameters for each type of supported dynamic remote storage.
    REMOTE_DYNAMIC_TYPE_CLASS_MAP: dict[str, StorageMetadataBase] = {
        REMOTE_DYNAMIC_TYPE_FILE:
            (FileStorageMetadata, set(inspect.getfullargspec(FileStorageMetadata).args) - {'self'}),
        REMOTE_DYNAMIC_TYPE_ORGANIZATION:
            (OrganizationStorageMetadata, set(inspect.getfullargspec(OrganizationStorageMetadata).args) - {'self'}),
        REMOTE_DYNAMIC_TYPE_ACTIVITY:
            (ActivityStorageMetadata, set(inspect.getfullargspec(ActivityStorageMetadata).args) - {'self'}),
        REMOTE_DYNAMIC_TYPE_USER:
            (UserStorageMetadata, set(inspect.getfullargspec(UserStorageMetadata).args) - {'self'}),
        REMOTE_DYNAMIC_TYPE_GROUP:
            (GroupStorageMetadata, set(inspect.getfullargspec(GroupStorageMetadata).args) - {'self'}),
        REMOTE_DYNAMIC_TYPE_RESOURCES_STATS:
            (ResourceConsumptionStorageMetadata, set(inspect.getfullargspec(ResourceConsumptionStorageMetadata).args) - {'self'}),
    }

    REMOTE_DYNAMIC_TYPE_SUPPORTED_LIST: set[str] = set(REMOTE_DYNAMIC_TYPE_CLASS_MAP.keys())

    # pylint: disable=too-many-branches,too-many-statements
    def __init__(self, *, lynceus_session: LynceusSession, lynceus_exchange: LynceusExchange | None,
                 lynceus_config: LynceusConfig = None,
                 env: str = None, env_suffix: str = None,
                 remote_config_section: str = None, remote_config_key: str = None,
                 remote_mode_forced_by_cli: bool = True, remote_mode_automatic_activation: bool = False,
                 source_path_format: str = '{target}/{env}/parquet', source_mode: bool = False,
                 dest_path_format: str = '{dest_file_name}/{source_name}.parquet', dest_path_kwargs: dict[str, str] = None):
        """
        Initializes Lynceus file Factory generating File allowing local or remote management,
         according to remote mode toggle (depending on specified argument and optional overriding configuration).
        :param lynceus_config: configuration to use (if not specified, configuration of specified lynceus session is used).
        :param env: name of the environment which will be used as parent directory of parquet files (for write access).
        :param remote_mode_forced_by_cli: (forced by CLI) True to read remotely, False to read locally (default).
        :param dest_path_format: the format used to generate destination path.
        :param dest_path_kwargs: the parameters used to generate destination path.
        :param remote_mode_automatic_activation: automatic activation requested.
        :param remote_config_section: config section containing remote storage configuration.
        :param remote_config_key: config key, in General section, giving config section (needed only if remote_config_section is not defined).

        remote_mode explanation:
         - by default destination files are read and write locally,
         - if Host is our Data Factory,
            or if NB_USER environment variable is defined to jovyan, indicating system is launched with Docker image,
            remote_mode automatic activation is requested
         - but the **override_to_local_mode** configuration (in configuration file), allow ignoring automatic activation request
         - in any case, the remote_mode_forced_by_cli toggle (e.g. set by a --remote-mode CLI option) can be used to force remote mode
            (it will be False here, BUT 100% of request will use the override_remote_mode method parameter).
         => this system could be lightened, but it is risky to do that while keeping backward compatibility.
        """
        super().__init__(logger_name='file', lynceus_session=lynceus_session, lynceus_exchange=lynceus_exchange)

        # Safe-guard:
        if not source_path_format:
            raise LynceusFileError('Source path format must be defined!')

        self.__env: str = env
        self.__remote_mode: bool = remote_mode_forced_by_cli
        self.__remote_mode_automatic_activation: bool = remote_mode_automatic_activation
        self.__local_mode_forced_by_config: bool = False
        self.__source_path_format: str = source_path_format
        self.__source_mode = source_mode
        self.__dest_path_format: str = dest_path_format
        self.__dest_path_kwargs: dict[str, str] = dest_path_kwargs or {}
        self.__previous_dest_path_kwargs = {}

        if not lynceus_config:
            lynceus_config = self._lynceus_session.get_lynceus_config_copy()

        # Loads remote configuration, and define some variable accordingly.
        self.__dynamic_container_name_params: dict[str, str] = {}
        if not remote_config_section and not remote_config_key:
            self._logger.warning('No remote configuration given at all, this file factory will only be able to create Local file.')
            self.__remote_config = None
            self.__with_dynamic_container_name: bool = False
        else:
            if not remote_config_section:
                remote_config_section = lynceus_config.get_config(CONFIG_GENERAL_KEY, remote_config_key)
            self.__remote_config = lynceus_config[remote_config_section]
            if not self.__remote_config:
                raise ValueError(f'There is no "{remote_config_section}" configuration in configuration file (are you sure you load the storage definition file ?).')

            self._logger.debug(f'According to "{remote_config_section}" configuration section, LynceusFileFactory will' +
                               f' consider remote configuration named "{remote_config_section}": "{LynceusConfig.format_config(self.__remote_config)}".')

            if self.__remote_config.get(CONFIG_STORAGE_REMOTE_TYPE) not in LynceusFileFactory.REMOTE_STORAGE_TYPE_SUPPORTED_LIST:
                raise NotImplementedError(f'Configured "{CONFIG_STORAGE_REMOTE_TYPE}={self.__remote_config.get(CONFIG_STORAGE_REMOTE_TYPE)}" is not supported.' +
                                          f' Supported values: {LynceusFileFactory.REMOTE_STORAGE_TYPE_SUPPORTED_LIST}')

            # Retrieves optional is_dynamic configuration option.
            self.__with_dynamic_container_name: bool = self.__remote_config.get(CONFIG_STORAGE_IS_DYNAMIC)
            self.__with_dynamic_container_type: str = self.__remote_config.get(CONFIG_STORAGE_DYNAMIC_TYPE,
                                                                               LynceusFileFactory.REMOTE_DYNAMIC_TYPE_GROUP)

            if self.__with_dynamic_container_type not in LynceusFileFactory.REMOTE_DYNAMIC_TYPE_SUPPORTED_LIST:
                raise NotImplementedError(f'Configured "{CONFIG_STORAGE_DYNAMIC_TYPE}={self.__with_dynamic_container_type}" is not supported.' +
                                          f' Supported values: {LynceusFileFactory.REMOTE_DYNAMIC_TYPE_SUPPORTED_LIST}')

            # Initializes some utilities.
            self.__s3utils = S3Utils(lynceus_session=lynceus_session, lynceus_exchange=lynceus_exchange, lynceus_s3_config=self.__remote_config)
            self.__s3utils.initialize()

            # Checks if local mode (against remote mode) is forced in configuration file.
            if 'override_to_local_mode' in self.__remote_config:
                if LynceusConfig.to_bool(self.__remote_config['override_to_local_mode']):
                    self.__local_mode_forced_by_config = True
                    # It is the case so defined the remote mode as False.
                    self.__remote_mode = False
                    # pylint: disable=logging-not-lazy
                    self._logger.info(
                        f'According to "override_to_local_mode" configuration in "{remote_config_section}"' +
                        ' remote mode is overridden to "local" (it can only be overridden by CLI option).')
            else:
                # In any case, set the remote mode as the value of CLI **or** auto activation.
                self.__remote_mode |= remote_mode_automatic_activation
                self._logger.info(f'remote mode="{self.__remote_mode}" (forced by CLI option="{remote_mode_forced_by_cli}";' +
                                  f' automatic activation according to environment="{self.__remote_mode_automatic_activation}").')

            if 'override_environment' in self.__remote_config:
                self.__env = self.__remote_config['override_environment']
                self.__env = self.__define_complete_env(self.__env, env_suffix)
                self._logger.info(f'According to "override_environment" configuration in "{remote_config_section}"' +
                                  f' Environnment is overriden to "{self.__env}".')

            # Defines remote root path, once for all.
            self.__remote_root_path: Path | None = None
            if not self.__with_dynamic_container_name:
                self.__remote_root_path = self.__define_remote_root_path(self.__env)

        # Defines default environment if needed.
        if self.__env is None:
            self.__env = self.__define_complete_env('dev', env_suffix)
            self._logger.info(f'No environment defined in CLI or configuration, defined it to "{self.__env}".')

        # Defines local root path, once for all.
        self.__local_root_path: Path = self.__define_local_root_path(self.__env)

        # Defines string presentation of this LynceusFile Factory.
        self.__string_presentation = LynceusConfig.format_dict_to_string(LynceusConfig.format_config(
            self.get_context_info() |
            {
                'env': self.__env,
                'source_path_format': self.__source_path_format,
                'source_mode': self.__source_mode,
                'dest_path_format': self.__dest_path_format,
                'storage': self.__remote_config or 'Local only',
                'dynamic': self.__dynamic_container_name_params,
            }), indentation_level=2)

    @property
    def is_dynamic_remote(self):
        return self.__with_dynamic_container_name

    def __define_complete_env(self, env: str, env_suffix: str):
        # Checks if this factory is used as a source.
        if self.__source_mode:
            # It is the case, so suffix is NOT used here.
            return env

        # It is used as a target, so environment suffix must be taken care.
        return env if not env_suffix else f'{env}/{env_suffix}'

    def force_cache_refresh(self):
        if self.__remote_config:
            self.__s3utils.force_cache_refresh()

    def get_env(self) -> str:
        return self.__env

    def __build_relative_path_dir(self, target: str, env: str | None):
        return self.__source_path_format.format(**self.__dest_path_kwargs, target=target, env=env)

    def __define_local_root_path(self, env: str | None) -> Path:
        root_path: Path = lookup_root_path('lynceus/misc', root_path=Path(__file__).parent)
        return root_path / Path(self.__build_relative_path_dir('target', env))

    def update_dynamic_storage_params(self, params: dict[str, str | int]):
        self.__dynamic_container_name_params: dict[str, str | int] = params.copy()
        mandatory_params: set[str] = {param for param, is_mandatory in CONFIG_PROJECT_DYNAMIC_STORAGE_MANDATORY_PARAM_MAP.items() if is_mandatory}

        # Ensures there are all the mandatory params.
        if mandatory_params - set(self.__dynamic_container_name_params.keys()):
            raise LynceusFileError(f'Specified dynamic storage params ({set(self.__dynamic_container_name_params.keys())}),' +
                                   f' should contain at least all the awaited ones ({mandatory_params}).')

        # Adds mandatory option allowing to create GroupStorageMetadata instance.
        # Tricks: GroupStorageMetadata is used only for its build_unique_storage_name method,
        #  which does NOT need following parameters, but which are mandatory to create a new instance.
        self.__dynamic_container_name_params.update({
            'organization_name': 'notUsedHere',
            'activity_name': 'notUsedHere',
            'topic_id': -1,
            'topic_name': 'notUsedHere',
            'group_order': -1,
            'group_nickname': 'notUsedHere',
        })

    def __define_remote_container_name(self) -> str:
        # Checks if it is a static or dynamic storage.
        if not self.__with_dynamic_container_name:
            return self.__remote_config["bucket_name"]

        # Defines which StorageMetadata and params according to configuration.
        storage_metadata_class, awaited_params = LynceusFileFactory.REMOTE_DYNAMIC_TYPE_CLASS_MAP.get(self.__with_dynamic_container_type)

        # Creates the corresponding StorageMetadata, and requests the unique storage name building to be 100% sure
        #  it will be the exact same name used during creation, and during compute resources request.
        try:
            # Filters parameters to use to instantiate such StorageMetadata class.
            params = {key: value for key, value in self.__dynamic_container_name_params.items()
                      if key in awaited_params}

            dynamic_storage: StorageMetadataBase = storage_metadata_class(**params)
            return dynamic_storage.build_unique_storage_name()
        except TypeError as exc:
            raise LynceusFileError(f'Unable to define the name of the dynamic remote container "{self}".', exc) from exc

    def __define_remote_root_path(self, env: str | None) -> Path:
        relative_path_dir: str = self.__build_relative_path_dir(self.__define_remote_container_name(), env)

        return Path(f'{LynceusFile.S3_PATH_BEGIN}{relative_path_dir}/')

    # pylint: disable=too-many-positional-arguments
    def new_file(self, source_name: str | None, source_file_name: str, must_exist: bool = True,
                 override_env: str = None, override_remote_mode: bool = None,
                 create_sub_directories: bool = True, dest_path_format: str = None,
                 override_dest_path_kwargs: dict = None, specific_dest_file_name: str = None) -> LynceusFile:
        # Safe-guard: ensures source_file_name is defined to minimum.
        source_file_name = source_file_name or '/'

        # Defines destination file name, which is the same as the source file name by default.
        dest_file_name = specific_dest_file_name if specific_dest_file_name else source_file_name

        # Manages path kwargs overriding if needed.
        path_kwargs = override_dest_path_kwargs if override_dest_path_kwargs else self.__dest_path_kwargs

        # Special Hack (mainly needed for CustomerInfo auto merge system), using previous path kwargs if none is defined here.
        if not path_kwargs:
            path_kwargs = self.__previous_dest_path_kwargs
        else:
            # Registers path_kwargs for next potential iteration.
            self.__previous_dest_path_kwargs = path_kwargs

        # Formats the new file path.
        if not dest_path_format:
            dest_path_format = self.__dest_path_format

        try:
            new_file_path: str = dest_path_format.format(**path_kwargs, source_name=source_name, dest_file_name=dest_file_name)
        except KeyError:
            # Gives as much information as possible.
            self._logger.error(f'Unable to build the file path from format "{dest_path_format}" and arguments: "{source_name}", "{dest_file_name}", "{path_kwargs=}"')
            # Stops on error anyway.
            raise

        return self._do_new_file(new_file_path, must_exist, override_env,
                                 override_remote_mode, create_sub_directories=create_sub_directories)

    # pylint: disable=too-many-positional-arguments
    def new_env_directory(self, must_exist: bool = True, override_env: str = None,
                          override_remote_mode: bool = None) -> LynceusFile:
        return self._do_new_file('', must_exist, override_env, override_remote_mode, create_sub_directories=False)

    def _do_new_file(self, path: str, must_exist: bool = True,
                     override_env: str | None = None, override_remote_mode: bool = None,
                     create_sub_directories: bool = True) -> LynceusFile:
        if self.__remote_mode or override_remote_mode:
            root_path: Path = self.__remote_root_path if override_env is None and not self.__with_dynamic_container_name \
                else self.__define_remote_root_path(override_env)

            complete_path: Path = root_path

            # Important: concatenates path only if it exists and not '/'.
            if path and path != '/':
                complete_path /= Path(path)

            # Manages optional globbing if needed.
            if '*' in path:
                matching_files = self.__s3utils.list_remote_files(remote_root_path=_RemoteS3LynceusFile.get_raw_path_from_remote_path(root_path), recursive=True, pattern=path, detail=True)
                if not matching_files:
                    self._logger.warning(f'Unable to find any remote files while globbing with "{complete_path}". It will certainly lead to not found file.')
                else:
                    # Sorts by last modification date.
                    sorted_matching_files = sorted(matching_files.items(), key=lambda kv: kv[1]['LastModified'], reverse=True)
                    selected_file_path = sorted_matching_files[0][0]

                    complete_path = root_path / Path(selected_file_path)
        else:
            root_path: Path = self.__local_root_path if override_env is None and not self.__with_dynamic_container_name \
                else self.__define_local_root_path(override_env)

            complete_path: Path = root_path / Path(path)
            # TODO: implement globbing on local path, something like root_path.glob(pattern)

        return self.create_from_full_path(complete_path, must_exist=must_exist,
                                          override_remote_mode=override_remote_mode, create_sub_directories=create_sub_directories)

    def create_from_full_path(self, complete_path: Path, must_exist: bool = True,
                              override_remote_mode: bool = None, create_sub_directories: bool = True) -> LynceusFile:
        complete_path = Path(complete_path)
        if self.__remote_mode or override_remote_mode:
            new_lynceus_file: LynceusFile = _RemoteS3LynceusFile(complete_path, self._logger,
                                                                 self.__s3utils.get_s3filesystem(),
                                                                 self.__s3utils)

            # Forces cache refresh if corresponding file existence if False atm.
            # it can happen if the file has been created (from elsewhere) after creation of this Factory.
            if must_exist and not new_lynceus_file.exists(reason='check if S3fs cache must be refreshed'):
                self.__s3utils.force_cache_refresh(path=_RemoteS3LynceusFile.get_raw_path_from_remote_path(complete_path.parent))
        else:
            # Creates subdirectories if needed.
            if create_sub_directories:
                complete_path.parent.mkdir(parents=True, exist_ok=True)

            new_lynceus_file: LynceusFile = _LocalLynceusFile(complete_path, self._logger)

        # Safe-guard: ensures corresponding file exists.
        if must_exist and not new_lynceus_file.exists():
            raise LynceusFileError(f'Requested file "{new_lynceus_file}" does not exist.')

        return new_lynceus_file

    def get_parent_file(self, lynceus_file: LynceusFile, must_exist: bool = True,
                        override_remote_mode: bool = None, create_sub_directories: bool = True) -> LynceusFile:
        return self.create_from_full_path(lynceus_file.get_parent_path(), must_exist=must_exist,
                                          override_remote_mode=override_remote_mode,
                                          create_sub_directories=create_sub_directories)

    def get_context_info(self):
        return {
            'remote_mode (from CLI)': self.__remote_mode,
            'remote_mode (automatic)': self.__remote_mode_automatic_activation,
            'local_mode (from Config)': self.__local_mode_forced_by_config
        }

    def __str__(self):
        return self.__string_presentation



import configparser
import logging.config
import logging.handlers
from logging import Logger
from pathlib import Path

from lynceus.core.config import CONFIG_GENERAL_KEY
from lynceus.core.config.lynceus_config import LynceusConfig
from lynceus.lynceus_exceptions import LynceusConfigError
from lynceus.utils import format_exception_human_readable, lookup_root_path


class LynceusSession:
    DEFAULT_SALT: str = 'lynceus'

    # Additional even more verbose than DEBUG log level.
    TRACE = 5

    __internal_key_preventing_external_usage_of_init = object()

    __REGISTERED_SESSIONS: dict = {}

    def __init__(self, *,
                 salt: str = DEFAULT_SALT,
                 root_logger_name: str = 'Lynceus',
                 load_core_default_config: bool = True,
                 overridden_root_logger_level: str | None = None,
                 _creation_key=None):
        """

        Creates a new Lynceus Session.

        Lynceus client must use :py:meth:`~lynceus.core.lynceus.LynceusSession.get_session` method only (it should NEVER call this constructor directly).

        :param salt: the salt used to identify with unicity this session.
        :param root_logger_name: the name of the root logger (will be used as prefix for all logger requested with :py:meth:`~lynceus.core.lynceus.LynceusSession.get_logger` method).
        :param load_core_default_config: toggle defining if Lynceus default configuration file must be loaded too (strongly recommended).
        :param overridden_root_logger_level: optional root logger level overridding; which can be interesting when used via a tool/script whose CLI options allow to force logging
                (i.e. --quiet, --debug ...)
        :param _creation_key: internal creation key (should NEVER be directly specified by external caller), preventing Lynceus client to call this method directly.
        """
        # Safe-guard: prevents direct call to this method.
        if _creation_key != LynceusSession.__internal_key_preventing_external_usage_of_init:
            raise RuntimeError(f'You should always use LynceusSession.get_session() method when requesting a LynceusSession ({salt=}).')

        # Initializes internal Lynceus config.
        self.__config = LynceusConfig()
        self.__root_logger_name = root_logger_name

        # Loads configuration file(s):
        #  - first the default config according to load_core_default_config toggle
        #  - then the default and user config corresponding to specify salt, if not the default one
        loaded_config_files = self.__load_configuration(salt, load_core_default_config=load_core_default_config)

        # Additional even more verbose than DEBUG log level.
        logging.TRACE = LynceusSession.TRACE
        logging.addLevelName(LynceusSession.TRACE, 'TRACE')

        # Setups default/root Lynceus logger.
        self.__logger = self.get_logger()
        root_level: int = logging.getLevelName(overridden_root_logger_level or self.get_config(CONFIG_GENERAL_KEY, 'logger.root.level', default='INFO'))
        self.__logger.setLevel(root_level)

        # Informs now that Logger is initialized.
        self.__logger.info(f'This is the loading information of looked up configuration files (Format=<Path: isLoaded?>) "{loaded_config_files}".')

    @staticmethod
    # pylint: disable=protected-access
    def get_session(*,
                    salt: str = DEFAULT_SALT,
                    registration_key: dict[str, str] | None = None,
                    root_logger_name: str = 'Lynceus',
                    load_core_default_config: bool = True,
                    overridden_root_logger_level: str | None = None):
        # Safe-guard: a registration key is highly recommended if not an internal session (which would be the case with salt == DEFAULT_SALT).
        if registration_key is None:
            if salt != LynceusSession.DEFAULT_SALT:
                raise LynceusConfigError('It is mandatory to provide your own registration_key when getting a session.' +
                                         f' System will automatically consider it as {registration_key}, but you may not be able to retrieve your session.' +
                                         ' And it may lead to memory leak.')
            registration_key = {'default': salt}

        # Turns the registration_key to an hashable version.
        if registration_key is not None:
            registration_key = frozenset(registration_key.items())

        # Creates and registers the session if needed.
        if registration_key in LynceusSession.__REGISTERED_SESSIONS:
            session: LynceusSession = LynceusSession.__REGISTERED_SESSIONS[registration_key]
            session.__logger.debug(f'Successfully retrieved cached Session for registration key "{registration_key}": {session}.')
        else:
            session: LynceusSession = LynceusSession(salt=salt, root_logger_name=root_logger_name,
                                                     load_core_default_config=load_core_default_config,
                                                     overridden_root_logger_level=overridden_root_logger_level,
                                                     _creation_key=LynceusSession.__internal_key_preventing_external_usage_of_init)

            LynceusSession.__REGISTERED_SESSIONS[registration_key] = session
            session.__logger.debug(f'Successfully registered new Session (new total count={len(LynceusSession.__REGISTERED_SESSIONS)})'
                                   f' for registration key "{registration_key}": {session}.'
                                   f' This is its loaded configuration (salt={salt}):\n' +
                                   f'{LynceusConfig.format_dict_to_string(LynceusConfig.format_config(session.__config), indentation_level=1)}')

        # TODO: implement another method allowing to free a specific registered session
        # TODO: turn LynceusSession as a Context, allowing usage with the 'with' keyword to automatic free the session once used

        # Returns the Lynceus session.
        return session

    def __load_configuration(self, salt: str, load_core_default_config: bool) -> dict[Path, bool]:
        # Defines the potential configuration file to load.
        config_file_meta_list = []
        if load_core_default_config:
            # Adds the Lynceus default configuration file as requested.
            config_file_meta_list.append({'config_path': f'misc/{self.DEFAULT_SALT}.default.conf', 'root_path': Path(__file__).parent})

        # Adds default configuration file corresponding to specified salt, if not the default one.
        if salt != self.DEFAULT_SALT:
            config_file_meta_list.append({'config_path': f'misc/{salt}.default.conf'})

        # Adds user configuration file corresponding to specified salt.
        config_file_meta_list.append({'config_path': f'{salt}.conf'})

        # Looks up for configuration file(s), starting with a "default" one, and then optional user one.
        loaded_config_files: dict[Path, bool] = {}
        for config_file_meta in config_file_meta_list:
            conf_file_name = config_file_meta['config_path']
            conf_file_root_path = config_file_meta.get('root_path', Path().resolve())

            loaded: bool = False
            relative_file: Path = Path(conf_file_name)
            try:
                full_path: Path = lookup_root_path(relative_file, root_path=conf_file_root_path) / relative_file

                # Merges it in internal Lynceus config.
                additional_config: configparser.ConfigParser = self.__config.update_from_configuration_file(full_path)

                # Updates/Configures fine-tuned logger, if any in this (first or) additional configuration file.
                # N.B.: in ideal world, could be interesting to extract logger config from all configuration file(s) merged, and
                #  create a fake fileConfig in memory, to logging.config.fileConfig once for all, after all loading.
                # Without that, handlers and formatters must be redefined each time, in each configuration file definined fine-tuned logger configuration.
                if 'loggers' in additional_config.sections():
                    self.get_logger('internal').debug(f'Requesting fine-tuned logger configuration with configuration coming from {conf_file_name}.')
                    # Cf. https://docs.python.org/3.10/library/logging.config.html#logging.config.fileConfig
                    logging.config.fileConfig(full_path, disable_existing_loggers=False, encoding='utf-8')

                loaded = True
            except FileNotFoundError as exc:
                # Since recent version, there is no more mandatory configuration file to have, and several files can be loaded.
                # N.B.: it is valid, there is no logger yet here, so using print ...
                self.get_logger('internal').debug(f'WARNING: Unable to load "{conf_file_name}" configuration file,'
                                                  f' from "{conf_file_root_path}"'
                                                  f' (it is OK, because this configuration file is NOT mandatory) =>'
                                                  f' {format_exception_human_readable(exc, quote_message=True)}')

            # Registers configuration file loading information.
            loaded_config_files[relative_file] = loaded

        # Returns configuration file loading mapping, for information.
        return loaded_config_files

    def has_config_section(self, section: str):
        return self.__config.has_section(section)

    def get_config(self, section: str, key: str, *, default: object = LynceusConfig.UNDEFINED_VALUE):
        return self.__config.get_config(section, key, default=default)

    def is_bool_config_enabled(self, section: str, key: str):
        return self.__config.is_bool_config_enabled(section=section, key=key)

    def get_config_section(self, section: str):
        return self.__config[section]

    def get_lynceus_config_copy(self) -> LynceusConfig:
        """
        Returns a copy of the internal Lynceus config, can be very interesting to get default config.
        :return: a copy of the internal Lynceus config, can be very interesting to get default config.
        """
        return self.__config.copy()

    def get_logger(self, name: str = None, *, parent_propagate: bool = True) -> Logger:
        complete_name: str = f'{self.__root_logger_name}.{name}' if name else self.__root_logger_name
        logger: Logger = logging.getLogger(complete_name)
        if logger.parent:
            logger.parent.propagate = parent_propagate

        return logger



# Defines various constants.
from pathlib import Path
from typing import Any

# * Date/time format
DATETIME_FORMAT = "%Y-%m-%dT%H:%M:%S.%f%z"
DATETIME_FORMAT_SHORT = "%Y-%m-%d"

# * Internals constants (avoiding circular import).
LYNCEUS_S3_CONFIG_KEY: str = 'lynceus_s3_config'

# * Configuration instance
CONFIG_GENERAL_KEY: str = 'General'

# Constants definition for 'Processing' configuration level: where any process can store information // used by jobs to share info.
CONFIG_PROCESS_KEY: str = 'Processing'
CONFIG_RUN_EXECUTION_ID: str = 'execution_id'

CONFIG_PROJECT_REPO_ANALYSE_DIR: str = 'uncompress_repo_directory'



CONFIG_PROCESS_REPO_ALL_FILE: str = 'interesting_files_list'
CONFIG_PROCESS_REPO_FILE_COUNT: str = 'interesting_files_count'

CONFIG_PROCESS_REPO_LINES_COUNT: str = 'interesting_lines_count'



# => Important: update this map each time a new statistic is added (which is very rare).
# The key is the statistic KEY, and the value is the optional transformation function.
CONFIG_PROCESS_ALL_STATISTICS_MAP: dict[str, Any] = {







}

# Constants definition for 'Storage' configuration level: where user can configure some directory to work into (like temporary root directory).
CONFIG_STORAGE_KEY: str = 'StorageConfiguration'


CONFIG_STORAGE_REMOTE_TYPE: str = 'storage_type'
CONFIG_STORAGE_LOCAL: str = 'Local'

CONFIG_STORAGE_IS_DYNAMIC: str = 'is_dynamic'
CONFIG_STORAGE_DYNAMIC_TYPE: str = 'dynamic_type'
CONFIG_STORAGE_REMOTE_AVAILABLE: str = 'available_remote_storages'
CONFIG_STORAGE_REMOTE_CONF_FILE: str = 'remote_storages_conf'


# Lynceus client (API, CLI ...) related config.
#  * configuration option
CONFIG_JSON_DUMP_KEY_END_KEYWORD: str = '_json_dump'
CONFIG_AUTHENTICATION_KEY: str = 'Authentication'







CONFIG_USER_ID: str = 'user_id'
CONFIG_PARTICIPANT_EMAIL: str = 'participant_email'
CONFIG_SCORE_PARTICIPANT: str = 'score_participant'
CONFIG_DISCRIMINANT: str = 'discriminant'
CONFIG_PROGRAM_ID: str = 'program_id'

#   - other parameters required for dynamic storage
CONFIG_PROJECT_DYNAMIC_STORAGE_H_INSTANCE_ID: str = 'instance_id'
CONFIG_PROJECT_DYNAMIC_STORAGE_H_ORGANIZATION_ID: str = 'organization_id'
CONFIG_PROJECT_DYNAMIC_STORAGE_H_ACTIVITY_ID: str = 'activity_id'
CONFIG_PROJECT_DYNAMIC_STORAGE_H_GROUP_ID: str = 'group_id'
#   - defines which dynamic storage are mandatory or not.
CONFIG_PROJECT_DYNAMIC_STORAGE_MANDATORY_PARAM_MAP: dict[str, bool] = {
    CONFIG_PROJECT_DYNAMIC_STORAGE_H_INSTANCE_ID: True,
    CONFIG_PROJECT_DYNAMIC_STORAGE_H_ORGANIZATION_ID: True,
    CONFIG_PROJECT_DYNAMIC_STORAGE_H_ACTIVITY_ID: True,
    CONFIG_PROJECT_DYNAMIC_STORAGE_H_GROUP_ID: True,
    CONFIG_USER_ID: False
}





#   - other parameters related to logging
CONFIG_LOG_MESSAGE_STATUS_KEY: str = 'message_status'

#  * various constants







# Constants definition for 'Jobs' configuration level: configuration of jobs.
CONFIG_JOBS_KEY: str = 'Jobs'
CONFIG_JOBS_DISABLED_LIST: str = 'disabled_job_list'




# Constants definition for 'Project' configuration level: configuration of project to analyze.
CONFIG_PROJECT_KEY: str = 'Project'
CONFIG_PROJECT_URI: str = 'uri'








CONFIG_PROJECT_CRED_SECRET: str = 'credential_secret'
CONFIG_PROJECT_CRED_SECRET_GITLAB_DEFAULT: str = 'gitlab_default_credential_secret'
CONFIG_PROJECT_CRED_SECRET_GITHUB_DEFAULT: str = 'github_default_credential_secret'




CONFIG_PROJECT_ROOT_PATH_HOLDER: str = '<PROJECT_ROOT_PATH>'

# Constants definition for 'Project profile metadata' configuration level: project profile metadata.









# * Job metadata definition.
# Job metadata keys.


# Job categories.








# * Scoring metadata definition.




# * Various configuration
# Constants definition for structure meta.











# CONFIG_STRUCTURE_CHECK_ALL_VALUES constant is interesting for Lynceus API to automatically remove any possible message prefix
#  when creating message with status.

                                                
                                                
                                                

GOOD_BAD_STRING: dict[bool, str] = {
    
    
    
}

FOUND_OR_NOT_STRING: dict[bool, str] = {
    
    
}

# * Compute Resources related constants.
ACCESS_PERMISSION_RO: str = 'RO'
ACCESS_PERMISSION_RW: str = 'RW'
ACCESS_PERMISSION_RWD: str = 'RWD'

# Keyword used when creating some computing resources.
CREATED_OBJECT_NAME_KEYWORD: str = 'aitraining'




from collections import defaultdict
from datetime import date, datetime, timedelta, timezone
from pathlib import Path

import github
import requests
from github import Github, GithubObject, Permissions, UnknownObjectException

from lynceus.core.exchange.lynceus_exchange import LynceusExchange
from lynceus.core.lynceus import LynceusSession
from lynceus.devops.devops_analyzer import DevOpsAnalyzer
from lynceus.utils import filter_kwargs
from lynceus.utils.lynceus_dict import LynceusDict


def github_exception_handler(func):
    def func_wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except github.GithubException as error:
            # Intercepts permission error.
            if error.status in (401, 403):
                raise PermissionError(
                    'You don\'t have enough permission to perform this operation on Github.') from error
            if error.status == 404:
                raise NameError('Unable to find requested Object.') from error

            # Raises any other error.
            raise

    return func_wrapper


# Cf. https://pygithub.readthedocs.io/en/latest/introduction.html
# Cf. https://docs.github.com/en/rest
# Cf. https://pygithub.readthedocs.io/en/stable/changes.html#breaking-changes for 2.x version breaking changes (mainly on datetime which are no more naive).
# Important:
#  - the Gitlab group notion corresponds to Organization notion on Github
#  - the Team notion of Github is NOT managed
class GithubDevOpsAnalyzer(DevOpsAnalyzer):
    __manager: Github

    def __init__(self, lynceus_session: LynceusSession, uri: str, token: str, lynceus_exchange: LynceusExchange):
        super().__init__(lynceus_session, uri, 'github', lynceus_exchange)
        kwargs = {'auth': github.Auth.Token(token), 'timeout': 60}
        if uri:
            kwargs['base_url'] = uri

        self.__manager = Github(**kwargs)

    # The following methods are only for uniformity and coherence.
    def _extract_user_info(self, user) -> LynceusDict:
        return LynceusDict({
            'id': user.id,
            'name': user.name,
            'login': user.login,
            'username': user.login,
            'e-mail': user.email,
            'avatar_url': user.avatar_url,
            'bio': user.bio
        })

    def _extract_group_info(self, group) -> LynceusDict:
        return LynceusDict({
            'id': group.id,
            'name': group.name,
            # N.B.: it does not seem to be any "path" for Github organization
            'path': group.login
        })

    def _extract_project_info(self, project) -> LynceusDict:
        return LynceusDict({
            'id': project.id,
            'name': project.name,
            'path': project.full_name,
            'web_url': project.html_url
        })

    def _extract_member_info(self, member) -> LynceusDict:
        # In Github the corresponding notion is 'contributors', and there is no 'status' information.
        return LynceusDict({
            'id': member.id,
            'name': member.name,
            'login': member.login,
            'username': member.login,
            'parent_id': self.INFO_UNDEFINED,
            'state': self.STATUS_ACTIVE
        })

    def _extract_issue_event_info(self, issue_event, **kwargs) -> LynceusDict:
        return LynceusDict({
                               'id': issue_event.id,
                               'issue_id': issue_event.issue.number,
                               'action': issue_event.event,
                               'target_type': 'issue',
                               'created_at': issue_event.created_at,
                               'author': self.INFO_UNDEFINED if not issue_event.actor else issue_event.actor.name,
                               'title': issue_event.issue.title,
                               'issue_web_url': kwargs['project_web_url'] + f'/issues/{issue_event.issue.number}'

                               # N.B.: project information is unable, and must be added by caller via kwargs.
                           } | kwargs)

    def _extract_commit_info(self, commit) -> LynceusDict:
        return LynceusDict({
            'id': commit.sha,
            'short_id': commit.sha[:8],
            'parent_ids': [parent_commit.sha for parent_commit in commit.parents],
            'message': commit.raw_data['commit'].get('message'),
            'created_at': commit.commit.author.date,
            'author_name': commit.author.name,
            'author_email': commit.raw_data['commit']['author']['email'],
            'committer_name': commit.committer.name,
            'committer_email': commit.raw_data['commit']['committer']['email'],
        })

    def _extract_branch_info(self, branch) -> LynceusDict:
        return LynceusDict({
            'name': branch.name,
            'merged': len(branch.commit.parents) > 1,
            'commit_id': branch.commit.sha,
            'commit_short_id': branch.commit.sha[:8],
            'created_at': branch.commit.commit.author.date,
        })

    def _extract_tag_info(self, tag) -> LynceusDict:
        return LynceusDict({
            'name': tag.name,
            'commit_id': tag.commit.sha,
            'commit_short_id': tag.commit.sha[:8],
            'created_at': tag.commit.commit.author.date,
        })

    # The following methods are only performing read access on DevOps backend.
    @github_exception_handler
    def authenticate(self):
        # There is no direct authentication with Github,
        #  simply retrieves the current user to test it.
        self._do_get_current_user()

    @github_exception_handler
    def _do_get_current_user(self):
        return self.__manager.get_user()

    # pylint: disable=unused-argument
    @github_exception_handler
    def _do_get_user_without_cache(self, *, username: str = None, email: str = None, **kwargs):
        kwargs_filtered = filter_kwargs(args_filter=['user_id'], **kwargs)
        if kwargs_filtered:
            return self.__manager.get_user_by_id(**kwargs_filtered)

        return self.__manager.get_user(login=username)

    @github_exception_handler
    def _do_get_groups(self, *, count: int | None = None, **kwargs):
        if count:
            return self.__manager.get_organizations(**kwargs)[:count]
        return self.__manager.get_organizations(**kwargs)

    @github_exception_handler
    def _do_get_group_without_cache(self, *, full_path: str, **kwargs):
        # https://pygithub.readthedocs.io/en/latest/github_objects/Repository.html?highlight=organization#github.Repository.Repository.organization
        # https://pygithub.readthedocs.io/en/latest/github_objects/Team.html?highlight=organization#github.Team.Team.organization
        # https://pygithub.readthedocs.io/en/latest/examples/MainClass.html?highlight=organization#get-organization-by-name

        try:
            # There are no optional kwargs available in this implementation.
            return self.__manager.get_organization(full_path)
        except UnknownObjectException as exception:
            raise NameError(
                f'Group "{full_path}" has not been found.') from exception

    @github_exception_handler
    def _do_get_projects(self, **kwargs):
        return self._do_get_current_user().get_repos(**kwargs)

    @github_exception_handler
    def _do_get_project_without_cache(self, *, full_path: str, **kwargs):
        try:
            # There are no optional kwargs available in this implementation.
            return self.__manager.get_repo(full_path)
        except UnknownObjectException as exception:
            raise NameError(
                f'Project/repository "{full_path}" has not been found.') from exception

    @github_exception_handler
    def check_permissions_on_project(self, *, full_path: str, get_metadata: bool, pull: bool,
                                     push: bool = False, maintain: bool = False, admin: bool = False, **kwargs):
        try:
            # Retrieves repository metadata (can lead to NameError if authenticated user has not enough permissions).
            repository = self._do_get_project(full_path=full_path, **kwargs)

            # From here, we consider get_metadata permission is OK.

            # Checks others permissions.
            permissions: Permissions = repository.permissions
            if pull and not permissions.pull:
                return False

            if push and not permissions.push:
                return False

            if maintain and not permissions.maintain:
                return False

            if admin and not permissions.admin:
                return False

            # All permissions check are OK.
            return True
        except (PermissionError, NameError):
            # Returns True if there were NO permission at all to check ...
            return not get_metadata and not pull and not push and not maintain and not admin

    @github_exception_handler
    def _do_get_project_commits(self, *, full_path: str, git_ref_name: str, count: int | None = None, **kwargs):
        repository = self._do_get_project(full_path=full_path, **kwargs)
        if count is not None:
            return repository.get_commits(sha=git_ref_name)[:count]

        return repository.get_commits(sha=git_ref_name)

    @github_exception_handler
    def _do_get_project_branches(self, *, full_path: str, **kwargs):
        repository = self._do_get_project(full_path=full_path, **kwargs)
        return repository.get_branches()

    @github_exception_handler
    def _do_get_project_tags(self, *, full_path: str, **kwargs):
        repository = self._do_get_project(full_path=full_path, **kwargs)

        return repository.get_tags()

    @github_exception_handler
    def _do_get_project_members(self, *, full_path: str, **kwargs):
        repository = self._do_get_project(full_path=full_path, **kwargs)
        return repository.get_contributors(**filter_kwargs(args_filter=['anon'], **kwargs))

    @github_exception_handler
    def _do_get_group_members(self, *, full_path: str, **kwargs):
        # https://pygithub.readthedocs.io/en/latest/github_objects/Organization.html?highlight=member#github.Organization.Organization.get_members
        organization = self._do_get_group(full_path=full_path, **kwargs)
        return organization.get_members(**kwargs)

    @github_exception_handler
    def _do_get_project_issue_events(self, *, full_path: str, action: str | None = None,
                                     from_date: datetime | None = None,
                                     to_date: datetime | None = None, **kwargs):
        # https://docs.github.com/en/developers/webhooks-and-events/events/github-event-types
        # https://docs.github.com/en/rest/activity/events

        repository = self._do_get_project(full_path=full_path, **kwargs)
        # Important: unfortunately it is NOT possible to filter on server side ...
        issue_events = repository.get_issues_events()

        # Filters issue event according to specified parameter(s).
        for issue_event in issue_events:
            if action and issue_event.event != action:
                continue

            if from_date and issue_event.created_at < from_date:
                # Stops the iteration since this issue event is the first being too old.
                break

            if to_date and issue_event.created_at > to_date:
                # Ignores this issue event.
                continue

            yield issue_event

    @github_exception_handler
    def _do_get_project_issues(self, *, full_path: str, **kwargs):
        repository = self._do_get_project(full_path=full_path, **kwargs)

        
        if 'state' not in kwargs:
            kwargs['state'] = 'all'

        return list(repository.get_issues(**kwargs))

    @github_exception_handler
    def _do_get_project_merge_requests(self, *, full_path: str, **kwargs):
        repository = self._do_get_project(full_path=full_path, **kwargs)

        
        if 'state' not in kwargs:
            kwargs['state'] = 'all'

        return list(repository.get_pulls(**kwargs))

    @github_exception_handler
    def _do_get_project_milestones(self, *, full_path: str, **kwargs):
        repository = self._do_get_project(full_path=full_path, **kwargs)
        return [milestone.title for milestone in repository.get_milestones(**kwargs)]

    @github_exception_handler
    def _do_get_group_milestones(self, *, full_path: str, **kwargs):
        self._logger.warning('There is milestone on Organization under Github (you need to check milestones directly from a project/repository).')
        return []

    @github_exception_handler
    def _do_get_group_projects(self, *, full_path: str, **kwargs):
        organization = self._do_get_group(full_path=full_path, **kwargs)
        # Removes potential 'recursive' parameter which is not supported here.
        if 'recursive' in kwargs:
            del kwargs['recursive']
        return organization.get_repos(**kwargs)

    @github_exception_handler
    def get_user_stats_commit_activity(self, *, group_full_path: str = None, project_full_path: str = None,
                                       since: datetime = None, keep_empty_stats: bool = False):
        # Cf. https://developer.github.com/v3/repos/statistics/#get-the-last-year-of-commit-activity-data

        # Defines threshold date.
        contributions_since: datetime = since if since else datetime.now(tz=timezone.utc) - timedelta(days=365)
        current_user = self._do_get_current_user()
        stats_user_commit_activity: dict[date, int] = defaultdict(int)
        for repo in current_user.get_repos():
            for commit_activity in (repo.get_stats_commit_activity() or []):
                # TODO: define how to ensure it is the good author ?
                # if not current_user.name == commit_activity.author:
                #     print('TODO: remove // AUTHOR is NOT the same: ', current_user.name, commit_activity.author)
                #     continue

                # Ignores oldest statistics.
                if commit_activity.week < contributions_since:
                    continue

                # Ignores 0 stats but if wanted.
                if not keep_empty_stats and not commit_activity.total:
                    continue

                # For each stats.
                for day, commit_count in enumerate(commit_activity.days):
                    # Ignores 0 stats but if wanted.
                    if not keep_empty_stats and not commit_count:
                        continue

                    day_date = commit_activity.week + timedelta(days=day)
                    stats_user_commit_activity[day_date.date()] += commit_count

        return stats_user_commit_activity

    @github_exception_handler
    def get_user_contributions(self, *, since: datetime = None, keep_empty_stats: bool = False):
        # Cf. https://developer.github.com/v3/repos/statistics/

        # Defines threshold date.
        contributions_since: datetime = since if since else datetime.now(tz=timezone.utc) - timedelta(days=365)
        current_user = self._do_get_current_user()
        stats_user_contributions: dict[date, dict[str, int]] = {}
        for repo in current_user.get_repos():
            for contribution in (repo.get_stats_contributors() or []):
                if not current_user.name == contribution.author:
                    continue

                # For each stats.
                for stats in contribution.weeks:
                    # Ignores oldest statistics.
                    if stats.w < contributions_since:
                        continue

                    # Ignores 0 stats but if wanted.
                    if not keep_empty_stats and not stats.a and not stats.d and not stats.c:
                        continue

                    try:
                        stats_map = stats_user_contributions[stats.w.date()]
                    except KeyError:
                        stats_map = {'additions': 0, 'deletions': 0, 'commits': 0}
                        stats_user_contributions[stats.w.date()] = stats_map

                    stats_map['additions'] += stats.a
                    stats_map['deletions'] += stats.d
                    stats_map['commits'] += stats.c

        return stats_user_contributions

    @github_exception_handler
    def get_user_stats_code_frequency(self):
        stats_code_frequency = {}
        for repo in self._do_get_current_user().get_repos():
            for stats in (repo.get_stats_code_frequency() or []):
                try:
                    stats_map = stats_code_frequency[stats.week]
                except KeyError:
                    stats_map = {'additions': 0, 'deletions': 0}
                    stats_code_frequency[stats.week] = stats_map

                stats_map['additions'] += stats.additions
                stats_map['deletions'] += stats.deletions
        return stats_code_frequency

    @github_exception_handler
    def _do_download_repository(self, *, project_full_path: str, dest_path: Path,
                                reference: str = None, chunk_size: int = 1024, **kwargs):
        # Cf. https://developer.github.com/v3/repos/contents/#get-archive-link
        # Cf. https://github.com/PyGithub/PyGithub/blob/main/github/Repository.py#L1535
        repository = self._do_get_project(full_path=project_full_path, **kwargs)
        url: str = repository.get_archive_link(archive_format='tarball', ref=reference or GithubObject.NotSet)
        with requests.get(url, stream=True, timeout=60) as reader:
            # Ensures specified reference exists.
            if reader.status_code == 404:
                raise NameError(f'Reference "{reference}" does not exists for project/repository "{project_full_path}".')

            with open(dest_path, 'wb') as dest_file:
                for chunk in reader.iter_content(chunk_size=chunk_size):
                    dest_file.write(chunk)

    # The following methods are performing write/delete operation on DevOps backend.
    @github_exception_handler
    def _do_create_group(self, *, parent_group_full_path: str, new_group_name: str, new_group_relative_path: str, **kwargs):
        raise NotImplementedError('This feature is not available with the implementation for Github.')

    @github_exception_handler
    def _do_create_user(self, *, name: str, username: str, email: str, **kwargs):
        raise NotImplementedError('This feature is not available with the implementation for Github.')

    @github_exception_handler
    def _do_update_user_notification_settings(self, *, username: str, email: str, **kwargs):
        raise NotImplementedError('This feature is not available with the implementation for Github.')

    @github_exception_handler
    def _do_add_group_member(self, *, group_full_path: str, username: str, access, **kwargs):
        # https://pygithub.readthedocs.io/en/latest/github_objects/Team.html?highlight=member#github.Team.Team.add_to_members
        raise NotImplementedError('This feature is not available with the implementation for Github.')

    @github_exception_handler
    def _do_export_project(self, *, project_full_path: str, export_dst_full_path: Path, timeout_sec: int = 60, **kwargs):
        raise NotImplementedError('This feature is not available with the implementation for Github.')

    @github_exception_handler
    def _do_import_project(self, *, parent_group_full_path: str,
                           new_project_name: str, new_project_path: str,
                           import_src_full_path: Path, timeout_sec: int = 60, **kwargs):
        raise NotImplementedError('This feature is not available with the implementation for Github.')



"""
Storage metadata classes allowing to easily configure various user, group and topic storages (default, and extra ones).

Inside Lynceus since version 0.7.0 (instead of LynceusCLI) to share unique key creation algorithm across all projects.

"""
from abc import ABC
from dataclasses import dataclass

from lynceus.core.config import ACCESS_PERMISSION_RO, ACCESS_PERMISSION_RW, ACCESS_PERMISSION_RWD, CREATED_OBJECT_NAME_KEYWORD
from lynceus.utils import cleansed_str_value

_MOUNT_HOME_RELPATH: str = ''

_MOUNT_ORGANIZATION_RELPATH_FORMAT: str = '/organization_{organization_id:05d}/{organization_name}_resources'
_MOUNT_EXTRA_USER_RELPATH_FORMAT: str = '/{activity_name}/user_{user_id:05d}_space'
_MOUNT_GROUP_RELPATH_FORMAT: str = '/{activity_name}/{topic_name}/group_{group_id_or_order:05d}_shared_workspace'
_MOUNT_GROUP_UNIQUE_RELPATH_FORMAT: str = '/{activity_name}/{topic_name}/group_shared_workspace'
_MOUNT_GROUP_SPECIAL_ORGANIZER_RELPATH_FORMAT: str = '/{activity_name}/organizers_shared_workspace'
_MOUNT_TOPIC_RELPATH_FORMAT: str = '/{activity_name}/{topic_name}/datasets'


@dataclass
class StorageMetadataBase(ABC):
    # TheInstance ID (allowing to have an environment discrimination)
    instance_id: str

    def cleansed_instance_id(self) -> str:
        return cleansed_str_value(self.instance_id)

    def build_unique_storage_name(self) -> str:
        raise NotImplementedError()

    def get_cache_toggle(self) -> bool:
        raise NotImplementedError()

    def get_access_permission(self) -> str:
        raise NotImplementedError()

    def build_storage_prefix(self) -> str:
        raise NotImplementedError()

    def build_mount_path_relative_to_home_dir(self) -> str:
        raise NotImplementedError()


@dataclass
class FileStorageMetadata(StorageMetadataBase):
    def build_unique_storage_name(self) -> str:
        return f'{self.cleansed_instance_id()}-core-api-file-s3-storage'

    def get_cache_toggle(self) -> bool:
        # This method should NOT be used on this kind of dynamic storage.
        raise NotImplementedError()

    def get_access_permission(self) -> str:
        # This method should NOT be used on this kind of dynamic storage.
        raise NotImplementedError()

    def build_storage_prefix(self) -> str:
        # This method should NOT be used on this kind of dynamic storage.
        raise NotImplementedError()

    def build_mount_path_relative_to_home_dir(self) -> str:
        # This method should NOT be used on this kind of dynamic storage.
        raise NotImplementedError()


@dataclass
class ResourceConsumptionStorageMetadata(StorageMetadataBase):
    def build_unique_storage_name(self) -> str:
        return f'{self.cleansed_instance_id()}-resources-stats-storage'

    def get_cache_toggle(self) -> bool:
        # This method should NOT be used on this kind of dynamic storage.
        raise NotImplementedError()

    def get_access_permission(self) -> str:
        # This method should NOT be used on this kind of dynamic storage.
        raise NotImplementedError()

    def build_storage_prefix(self) -> str:
        # This method should NOT be used on this kind of dynamic storage.
        raise NotImplementedError()

    def build_mount_path_relative_to_home_dir(self) -> str:
        # This method should NOT be used on this kind of dynamic storage.
        raise NotImplementedError()


@dataclass
class OrganizationStorageMetadata(StorageMetadataBase):
    organization_id: int
    organization_name: str

    def cleansed_organization_name(self) -> str:
        return cleansed_str_value(self.organization_name)

    def build_unique_storage_name(self) -> str:
        return f'{self.cleansed_instance_id()}-{self.organization_id:05d}-org-{CREATED_OBJECT_NAME_KEYWORD}'

    def get_cache_toggle(self) -> bool:
        return True

    def get_access_permission(self) -> str:
        # Organization storages contents are only managed via client project, so here it is only RO.
        return ACCESS_PERMISSION_RO

    def build_storage_prefix(self) -> str:
        # No prefix for this type of storage.
        return ''

    def build_mount_path_relative_to_home_dir(self) -> str:
        return _MOUNT_ORGANIZATION_RELPATH_FORMAT.format(
            organization_id=self.organization_id,
            organization_name=self.cleansed_organization_name()
        )


@dataclass
class StorageMetadataWithActivity(StorageMetadataBase, ABC):
    # TheActivity ID, and name.
    activity_id: int
    activity_name: str

    def cleansed_activity_name(self) -> str:
        return cleansed_str_value(self.activity_name)


@dataclass
class UserStorageMetadata(StorageMetadataWithActivity):
    # TheUser ID.
    user_id: int

    # Defines if write permission is allowed on this storage when NOT the default one (RO by default for extra user storage).
    write_permission: bool = False

    # Defines if Cache must be forced (can be interesting when participant are using multiple instance for the same Activity, for instance cpu/gpu).
    force_cache: bool = False

    # Default toggle: if True, it corresponds to the default user/group/topic space (changes behaviour on mount point and/or permissions).
    default: bool = False

    def build_unique_storage_name(self) -> str:
        # Important: see this Gitlab Card to see before/after algorithm:
        #  

        return f'{self.cleansed_instance_id()}-{self.user_id:05d}-user-{CREATED_OBJECT_NAME_KEYWORD}'

    def get_cache_toggle(self) -> bool:
        # According to situation, if it is the default/contextual user storage or an extra mounted one:
        #  - default/contextual => no need for cache
        #  - extra => interesting to have the cache
        return self.force_cache or not self.default

    def get_access_permission(self) -> str:
        # According to situation, if it is the default/contextual user storage or an extra mounted one:
        #  - default/contextual => user has RW access on its own user storage
        #  - extra => RO by default, RW only if defined
        return ACCESS_PERMISSION_RWD if self.default or self.write_permission else ACCESS_PERMISSION_RO

    def build_storage_prefix(self) -> str:
        # No prefix for this type of storage.
        return ''

    def build_mount_path_relative_to_home_dir(self) -> str:
        # Checks if it is the default User storage, in which case it must be mounted as $HOME.
        if self.default:
            return _MOUNT_HOME_RELPATH

        return _MOUNT_EXTRA_USER_RELPATH_FORMAT.format(
            activity_name=self.cleansed_activity_name(),
            user_id=self.user_id
        )


@dataclass
class StorageMetadataWithTopicBase(StorageMetadataWithActivity, ABC):
    # TheTopic ID and name.
    topic_id: int
    topic_name: str

    def cleansed_topic_name(self) -> str:
        return cleansed_str_value(self.topic_name)


@dataclass
# pylint: disable = too-many-instance-attributes
class GroupStorageMetadata(StorageMetadataWithTopicBase):
    # TheGroup metadata.
    group_id: int
    group_order: int
    group_nickname: str

    # Defines if write permission is allowed on this storage when NOT the default one (RO by default for extra group storage).
    write_permission: bool = False

    # Defines if Delete permission is allowed (false by default since 2023 January).
    # Requires write_permission True to be considered.
    delete_permission: bool = False

    # Toggle: defines if the mount point should contain group id (it allows to have a proper mount point name, in case there is no group, or only one group).
    mount_point_with_group_id: bool = True

    # Toggle: defines to use the group order (which is known by human) in the mount point, instead of the group id (which is only internal)
    use_group_order_instead_of_group_id: bool = False

    # Toggle: defines if this is a special organizer storage.
    special_organizer_group_storage: bool = False

    # Default toggle: if True, it corresponds to the default user/group/topic space (changes behaviour on mount point and/or permissions).
    default: bool = False

    def build_unique_storage_name(self) -> str:
        # Important: see this Gitlab Card to see before/after algorithm:
        #  

        return f'{self.cleansed_instance_id()}-{self.activity_id:05d}-{self.group_id:05d}-group-{CREATED_OBJECT_NAME_KEYWORD}'

    def build_mount_path_relative_to_home_dir(self) -> str:
        if self.special_organizer_group_storage:
            return _MOUNT_GROUP_SPECIAL_ORGANIZER_RELPATH_FORMAT.format(
                activity_name=self.cleansed_activity_name()
            )

        if self.mount_point_with_group_id:
            final_id = self.group_order if self.use_group_order_instead_of_group_id else self.group_id

            return _MOUNT_GROUP_RELPATH_FORMAT.format(
                activity_name=self.cleansed_activity_name(),
                topic_name=self.cleansed_topic_name(),
                group_id_or_order=final_id
            )

        return _MOUNT_GROUP_UNIQUE_RELPATH_FORMAT.format(
            activity_name=self.cleansed_activity_name(),
            topic_name=self.cleansed_topic_name()
        )

    def get_cache_toggle(self) -> bool:
        # In any case Cache is needed on Group storage.
        return True

    def get_access_permission(self) -> str:
        # According to situation, if it is the default/contextual group storage or an extra mounted one:
        #  - default/contextual => user has RW acces on its own group storage
        #  - extra => RO by default, RW only if defined
        #  - in any case, RW becomes RWD if delete_permission is True
        if self.default or self.write_permission:
            return ACCESS_PERMISSION_RWD if self.delete_permission else ACCESS_PERMISSION_RW

        return ACCESS_PERMISSION_RO

    def build_storage_prefix(self) -> str:
        return f'{self.activity_id:05d}/'


@dataclass
class ActivityStorageMetadata(StorageMetadataWithTopicBase):
    # Defines if write permission is allowed on this storage when NOT the default one (RO by default for extra group storage).
    write_permission: bool = False

    # Defines if Delete permission is allowed (false by default since 2023 January).
    # Requires write_permission True to be considered.
    delete_permission: bool = False

    # Defines if access should be given at root of the topic/fact/datasets storage (not recommended but for Ubeeko Team member)
    force_topic_storage_root: bool = False

    def build_unique_storage_name(self) -> str:
        return f'{self.cleansed_instance_id()}-{self.activity_id:05d}-activity-{CREATED_OBJECT_NAME_KEYWORD}'

    def build_mount_path_relative_to_home_dir(self) -> str:
        return _MOUNT_TOPIC_RELPATH_FORMAT.format(
            activity_name=self.cleansed_activity_name(),
            topic_name=self.cleansed_topic_name(),
            topic_id=self.topic_id
        )

    def get_cache_toggle(self) -> bool:
        # Important: cache allows sharing (the same Ceph) all the same files for all instances mounting this share.
        # So it is very important to use cache on Topic/fact/datasets storage for OVH to mount it only once for any count of AI Training instance(s) ...
        return True

    def get_access_permission(self) -> str:
        # According to situation:
        #  - if no write permission => RO
        #  - if ONLY write permission => RW
        #  - in write AND delete permission => RWD
        if not self.write_permission:
            return ACCESS_PERMISSION_RO

        return ACCESS_PERMISSION_RWD if self.delete_permission else ACCESS_PERMISSION_RW

    def build_storage_prefix(self) -> str:
        return f'{self.cleansed_instance_id()}/' + \
               f'{self.activity_id:05d}/{self.topic_id:05d}/public/' if not self.force_topic_storage_root else ''


# Backward compatibility, ActivityStorageMetadata was formerly named TopicStorageMetadata when is was related to topic
TopicStorageMetadata = ActivityStorageMetadata



import random
import string








def random_int(start: int = 1, stop: int = 424242) -> int:
    return random.randint(start, stop)


def random_id(start: int = 1, stop: int = 424242) -> int:
    return random_int(start=start, stop=stop)


def random_bool() -> bool:
    return random_id() % 2 == 0


def random_string(size: int = 8, *, prefix: str = '', population=string.ascii_letters) -> str:
    return prefix + ''.join(random.choices(population=population, k=size))


def random_email():
    first_name: str = random_string()
    last_name: str = random_string()
    domain: str = f'{random_string(size=5)}.{random_string(size=3)}'
    return f'{first_name}.{last_name}@{domain}'


def random_path(*, part_size: int = 4, part_count: int = 3) -> str:
    path_parts = []
    for _ in range(part_count):
        path_parts.append(random_string(part_size))

    return '/'.join(path_parts)


def random_password(size: int = 16) -> str:
    return random_string(size, population=string.printable)



    # pylint: disable=not-an-iterable







def random_enum(enum_class):
    return random.choice([item.value for item in enum_class])











import logging
from logging import StreamHandler


class FilteredStdoutLoggerHandler(StreamHandler):
    def __init__(self, stream=None):
        super().__init__(stream=stream)

        # Adds automatically filter to NOT log messages from Warning to more, which should be logged on stderr, and not stdout to avoid duplicates.
        self.addFilter(lambda record: record.levelno < logging.WARNING)



from datetime import datetime
from pathlib import Path

from lynceus.core.exchange.lynceus_exchange import LynceusExchange
from lynceus.core.lynceus import LynceusSession
from lynceus.core.lynceus_client import LynceusClientClass
from lynceus.utils.lynceus_dict import LynceusDict


# TODO: add members?
# TODO: add accessrequests/permissions listing feature


# pylint: disable=too-many-public-methods
class DevOpsAnalyzer(LynceusClientClass):
    STATUS_ACTIVE: str = 'active'
    INFO_UNDEFINED: str = 'undefined'

    CACHE_USER_TYPE: str = 'user'
    CACHE_GROUP_TYPE: str = 'group'
    CACHE_PROJECT_TYPE: str = 'project'

    def __init__(self, lynceus_session: LynceusSession, uri: str, logger_name: str, lynceus_exchange: LynceusExchange):
        super().__init__(lynceus_session, logger_name, lynceus_exchange)
        self._uri = uri

        # Initializes cache system.
        self.__cache = {
            self.CACHE_USER_TYPE: {},
            self.CACHE_GROUP_TYPE: {},
            self.CACHE_PROJECT_TYPE: {},
        }

    # The following methods are only for uniformity and coherence.
    def _extract_user_info(self, user) -> LynceusDict:
        raise NotImplementedError()

    def _extract_group_info(self, group) -> LynceusDict:
        raise NotImplementedError()

    def _extract_project_info(self, project) -> LynceusDict:
        raise NotImplementedError()

    def _extract_member_info(self, member) -> LynceusDict:
        raise NotImplementedError()

    def _extract_issue_event_info(self, issue_event, **kwargs) -> LynceusDict:
        raise NotImplementedError()

    def _extract_commit_info(self, commit) -> LynceusDict:
        raise NotImplementedError()

    def _extract_branch_info(self, branch) -> str:
        raise NotImplementedError()

    def _extract_tag_info(self, tag) -> str:
        raise NotImplementedError()

    def __get_from_cache(self, *, cache_type: str, cache_key, log_access: bool = True):
        if log_access:
            self._logger.debug(f'Checking if an instance is registered in cache, for type "{cache_type}" and key "{cache_key}".')
        return self.__cache[cache_type].get(cache_key, None)

    def __register_in_cache(self, *, cache_type: str, cache_key, obj, obj_short):
        # Safe-guard: checks if it has already been registered in cache.
        from_cache = self.__get_from_cache(cache_type=cache_type, cache_key=cache_key, log_access=False)
        if from_cache:
            self._logger.warning(f'An instance of type "{cache_type}" has already been registered in cache, for key "{cache_key}". It will be overridden.')

        self._logger.debug(f'Registering a complete/long instance of type "{cache_type}" in cache, for key "{cache_key}", whose short version is: "{obj_short}".')
        self.__cache[cache_type][cache_key] = obj

    # The following methods are only performing read access on DevOps backend.
    def authenticate(self):
        raise NotImplementedError()

    def _do_get_current_user(self):
        raise NotImplementedError()

    def get_current_user(self):
        self._logger.debug('Retrieving current user information.')
        user = self._do_get_current_user()

        return self._extract_user_info(user)

    def _do_get_user_without_cache(self, *, username: str = None, email: str = None, **kwargs):
        raise NotImplementedError()

    def _do_get_user(self, *, username: str = None, email: str = None, **kwargs):
        # Checks if available in cache.
        cache_key = (username, email)
        user = self.__get_from_cache(cache_type=self.CACHE_USER_TYPE, cache_key=cache_key)

        # Retrieves it if not available in cache.
        if not user:
            user = self._do_get_user_without_cache(username=username, email=email, **kwargs)

            # Registers in cache.
            self.__register_in_cache(cache_type=self.CACHE_USER_TYPE, cache_key=cache_key,
                                     obj=user, obj_short=self._extract_user_info(user))

        return user

    def get_user(self, *, username: str = None, email: str = None, **kwargs):
        self._logger.debug(f'Retrieving user "{username=}" with "{email}" ({kwargs=}).')
        user = self._do_get_user(username=username, email=email, **kwargs)
        return self._extract_user_info(user)

    def _do_get_groups(self, *, count: int | None = None, **kwargs):
        raise NotImplementedError()

    def get_groups(self, *, count: int | None = None, **kwargs):
        self._logger.debug(f'Retrieving groups ({count=}; {kwargs=}).')
        groups = self._do_get_groups(count=count, **kwargs)
        return [self._extract_group_info(group) for group in groups]

    def _do_get_group_without_cache(self, *, full_path: str, **kwargs):
        raise NotImplementedError()

    def _do_get_group(self, *, full_path: str, **kwargs):
        # Checks if available in cache.
        cache_key = (full_path,)
        group = self.__get_from_cache(cache_type=self.CACHE_GROUP_TYPE, cache_key=cache_key)

        # Retrieves it if not available in cache.
        if not group:
            group = self._do_get_group_without_cache(full_path=full_path, **kwargs)

            # Registers in cache.
            self.__register_in_cache(cache_type=self.CACHE_GROUP_TYPE, cache_key=cache_key,
                                     obj=group, obj_short=self._extract_group_info(group))

        return group

    def get_group(self, *, full_path: str, **kwargs):
        self._logger.debug(f'Retrieving group "{full_path=}" ({kwargs=}).')
        group = self._do_get_group(full_path=full_path, **kwargs)
        return self._extract_group_info(group)

    def _do_get_projects(self, **kwargs):
        raise NotImplementedError()

    def get_projects(self, **kwargs):
        self._logger.debug(f'Retrieving projects ({kwargs=}).')
        projects = self._do_get_projects(**kwargs)
        return [self._extract_project_info(project) for project in projects]

    def _do_get_project_without_cache(self, *, full_path: str, **kwargs):
        raise NotImplementedError()

    def _do_get_project(self, *, full_path: str, **kwargs):
        # Checks if available in cache.
        cache_key = (full_path,)
        project = self.__get_from_cache(cache_type=self.CACHE_PROJECT_TYPE, cache_key=cache_key)

        # Retrieves it if not available in cache.
        if not project:
            project = self._do_get_project_without_cache(full_path=full_path, **kwargs)

            # Registers in cache.
            self.__register_in_cache(cache_type=self.CACHE_PROJECT_TYPE, cache_key=cache_key,
                                     obj=project, obj_short=self._extract_project_info(project))

        return project

    def get_project(self, *, full_path: str, **kwargs):
        self._logger.debug(f'Retrieving project "{full_path=}" ({kwargs=}).')
        project = self._do_get_project(full_path=full_path, **kwargs)
        return self._extract_project_info(project)

    def check_permissions_on_project(self, *, full_path: str, get_metadata: bool, pull: bool,
                                     push: bool = False, maintain: bool = False, admin: bool = False, **kwargs):
        """
        Checks permission of Authenticated user on specific project (full_path and optional kwargs).

        :param full_path: full path of the project (including namespace)
        :param get_metadata: check 'get project metadata' permission.
        :param pull: check 'pull' permission.
        :param push: check 'push' permission (does not check this permission by default).
        :param maintain: check 'maintain' permission (does not check this permission by default).
        :param admin: check 'admin' permission (does not check this permission by default).
        :param kwargs: optional additional parameters
        :return: True if the Authenticated user has requested permission, False otherwise.
        """
        raise NotImplementedError()

    def _do_get_project_members(self, *, full_path: str, **kwargs):
        raise NotImplementedError()

    def get_project_commits(self, *, full_path: str, git_ref_name: str, count: int | None = None, **kwargs):
        self._logger.debug(f'Retrieving commits from git_reference "{git_ref_name}" of project "{full_path}".')
        commits = self._do_get_project_commits(full_path=full_path, git_ref_name=git_ref_name, count=count, **kwargs)
        return [self._extract_commit_info(commit) for commit in commits]

    def _do_get_project_commits(self, *, full_path: str, git_ref_name: str, count: int | None = None, **kwargs):
        raise NotImplementedError()

    def get_project_branches(self, *, full_path: str, **kwargs):
        self._logger.debug(f'Retrieving branches of project "{full_path=}" ({kwargs=}).')
        branches = self._do_get_project_branches(full_path=full_path, **kwargs)
        return [self._extract_branch_info(branch) for branch in branches]

    def _do_get_project_branches(self, *, full_path: str, **kwargs):
        raise NotImplementedError()

    def get_project_tags(self, *, full_path: str, **kwargs):
        self._logger.debug(f'Retrieving tags of project "{full_path=}" ({kwargs=}).')
        tags = self._do_get_project_tags(full_path=full_path, **kwargs)
        return [self._extract_tag_info(tag) for tag in tags]

    def _do_get_project_tags(self, *, full_path: str, **kwargs):
        raise NotImplementedError()

    def get_project_members(self, *, full_path: str, **kwargs):
        """
        Return all members of project.

        :param full_path: full path of the project (including namespace)
        :return: all members of the project
        """
        self._logger.debug(f'Retrieving members of project "{full_path=}" ({kwargs=}).')
        project_members = self._do_get_project_members(full_path=full_path, **kwargs)
        return [self._extract_member_info(member) for member in project_members]

    def _do_get_group_members(self, *, full_path: str, **kwargs):
        raise NotImplementedError()

    def get_group_members(self, *, full_path: str, **kwargs):
        """
        Return all members of group (aka Organization).

        :param full_path: full path of the group (including namespace)
        :return: all members of the group
        """
        self._logger.debug(f'Retrieving members of group "{full_path=}" ({kwargs=}).')
        members = self._do_get_group_members(full_path=full_path, **kwargs)
        return [self._extract_member_info(member) for member in members]

    def _do_get_project_issue_events(self, *, full_path: str, action: str | None = None,
                                     from_date: datetime | None = None,
                                     to_date: datetime | None = None, **kwargs):
        raise NotImplementedError()

    def get_project_issue_events(self, *, full_path: str, action: str | None = None,
                                 from_date: datetime | None = None,
                                 to_date: datetime | None = None, **kwargs):
        """
        Returns project issue events (filtered according to specified parameters).

        :param full_path: path of the project (including namespace)
        :param action: optional issue event action to filter on, most of them are common to DevOps
                See this one for Gitlab specs: https://docs.gitlab.com/ee/user/profile/index.html#user-contribution-events
                See this one for Github specs: https://docs.github.com/en/developers/webhooks-and-events/events/github-event-types#event-payload-object-6
        :param from_date: optional datetime from which to consider issue events
        :param to_date: optional datetime until which to consider issue events
        :param kwargs: optional additional parameters
        :return: filtered issue events of the project.
        """
        self._logger.debug(f'Retrieving issue events of project "{full_path=}" ({action=}; {from_date=}; {to_date=}; {kwargs=}).')
        project_events = self._do_get_project_issue_events(full_path=full_path, action=action, from_date=from_date, to_date=to_date, **kwargs)

        # Important: in Github there is no project (either id or name) information attached to event, and on Gitlab there is only the id.
        # To return consistent result, we add both here.
        project = self.get_project(full_path=full_path)
        project_metadata = {
            'project_id': project.id,
            'project_name': project.name,
            'project_web_url': project.web_url,
        }

        return [self._extract_issue_event_info(event, **project_metadata) for event in project_events]

    def _do_get_project_issues(self, *, full_path: str, **kwargs):
        raise NotImplementedError()

    def get_project_issues(self, *, full_path: str, **kwargs):
        """
        Return all issues of project.

        :param full_path: path of the project (including namespace)
        :return: all issues of the project
        """
        self._logger.debug(f'Retrieving issues of project "{full_path=}" {kwargs=}).')
        return self._do_get_project_issues(full_path=full_path, **kwargs)

    def get_project_pull_requests(self, *, full_path: str, **kwargs):
        return self.get_project_merge_requests(full_path=full_path, **kwargs)

    def _do_get_project_merge_requests(self, *, full_path: str, **kwargs):
        raise NotImplementedError()

    def get_project_merge_requests(self, *, full_path: str, **kwargs):
        """
        Return all issues of project.

        :param full_path: path of the project (including namespace)
        :return: all issues of the project
        """
        self._logger.debug(f'Retrieving merge/pull requests of project "{full_path=}" {kwargs=}).')
        return self._do_get_project_merge_requests(full_path=full_path, **kwargs)

    def _do_get_project_milestones(self, *, full_path: str, **kwargs):
        raise NotImplementedError()

    def get_project_milestones(self, *, full_path: str, **kwargs):
        """
        Return all milestones (aka sprints) of project.

        :param full_path: path of the project (including namespace)
        :return: all milestones of the project
        """
        self._logger.debug(f'Retrieving milestones of project "{full_path=}" {kwargs=}).')
        return self._do_get_project_milestones(full_path=full_path, **kwargs)

    def _do_get_group_projects(self, *, full_path: str, **kwargs):
        """
        Return all projects of group.

        :param full_path: path of the group
        :return: all projects of the group
        """
        raise NotImplementedError()

    def get_group_projects(self, *, full_path: str, **kwargs):
        self._logger.debug(f'Retrieving projects of group "{full_path=}" ({kwargs=}).')
        projects = self._do_get_group_projects(full_path=full_path, **kwargs)
        return [self._extract_project_info(project) for project in projects]

    def _do_get_group_milestones(self, *, full_path: str, **kwargs):
        raise NotImplementedError()

    def get_group_milestones(self, *, full_path: str, **kwargs):
        """
        Return all milestones (aka sprints) of group.

        :param full_path: path of the group
        :return: all milestones of the group
        """
        self._logger.debug(f'Retrieving milestones of group "{full_path=}" {kwargs=}).')
        return self._do_get_group_milestones(full_path=full_path, **kwargs)

    def get_user_stats_commit_activity(self, *, group_full_path: str = None, project_full_path: str = None,
                                       since: datetime = None, keep_empty_stats: bool = False):

        """
        Computes total statistics of all accessible projects of authenticated user, or all projects of specified group
         and/or of specified project.
        :param group_full_path: (optional) statistics will be computed on all projects of this group.
        :param project_full_path: (optional) statistics will be computed on this project.
        :param since: start date from which statistics must be computed.
        :param keep_empty_stats:
        """
        raise NotImplementedError()

    def get_user_contributions(self, *, since: datetime = None, keep_empty_stats: bool = False):
        raise NotImplementedError()

    def get_user_stats_code_frequency(self):
        raise NotImplementedError()

    def download_repository(self, *, project_full_path: str, dest_path: Path, reference: str = None,
                            chunk_size: int = 1024, **kwargs):
        self._logger.debug(f'Starting repository download of project "{project_full_path}" with git reference "{reference}", to destination file "{dest_path}" ...')
        try:
            self._do_download_repository(project_full_path=project_full_path, dest_path=dest_path, reference=reference, chunk_size=chunk_size, **kwargs)
            self._logger.info(f'Successfully downloaded repository of project "{project_full_path}" with git reference "{reference}", to destination file "{dest_path}".')
        except NameError as exc:
            
            
            #  like with hacked wrapper/2022-wrapper-lynceus-test.conf
            git_reference_str: str = f'git reference "{reference}"' if reference is not None else 'default git reference'
            error_message: str = f'Unable to download/access {git_reference_str} of project "{project_full_path}" (ensure your Token has permissions enough).'
            self._logger.warning(error_message)
            raise ValueError(error_message) from exc

    def _do_download_repository(self, *, project_full_path: str, dest_path: Path, reference: str = None,
                                chunk_size: int = 1024, **kwargs):
        raise NotImplementedError()

    # The following methods are performing write/delete operation on DevOps backend.
    def _do_create_group(self, *, parent_group_full_path: str, new_group_name: str, new_group_relative_path: str, **kwargs):
        """
        Create a new group.

        :param parent_group_full_path: full path of the parent group in which to create this new group, None to create a new root group.
        :param new_group_name: name of the group to create.
        :param new_group_relative_path: path of the group to create, relative to its optional parent group.
        :param kwargs: any optional parameters (e.g. visibility, project_creation_level ...)
        :return: an usable presentation of the created group.
        """
        raise NotImplementedError()

    def create_group(self, *, parent_group_full_path: str, new_group_name: str, new_group_relative_path: str, **kwargs):
        self._logger.debug(f'Creating group "{new_group_name=}" ({parent_group_full_path=}; {new_group_relative_path=}; {kwargs=}).')
        group = self._do_create_group(parent_group_full_path=parent_group_full_path, new_group_name=new_group_name,
                                      new_group_relative_path=new_group_relative_path, **kwargs)
        return self._extract_group_info(group)

    def _do_create_user(self, *, name: str, username: str, email: str, **kwargs):
        raise NotImplementedError()

    def create_user(self, *, name: str, username: str, email: str, **kwargs):
        self._logger.debug(f'Creating user "{name=}" ({username=}; {email=}; {kwargs=}).')
        user = self._do_create_user(name=name, username=username, email=email, **kwargs)
        return self._extract_user_info(user)

    def _do_update_user_notification_settings(self, *, username: str, email: str, **kwargs):
        raise NotImplementedError()

    def update_user_notification_settings(self, *, username: str, email: str, **kwargs):
        self._logger.debug(f'Updating notification settings for "{username=}" ({email=}; {kwargs=}).')
        user = self._do_update_user_notification_settings(username=username, email=email, **kwargs)
        return self._extract_user_info(user)

    def _do_add_group_member(self, *, group_full_path: str, username: str, access, **kwargs):
        raise NotImplementedError()

    def add_group_member(self, *, group_full_path: str, username: str, access, **kwargs):
        self._logger.debug(f'Adding user "{username=}" as member of group "{group_full_path=} ({access=}; {kwargs=}).')
        member = self._do_add_group_member(group_full_path=group_full_path, username=username, access=access, **kwargs)
        return self._extract_member_info(member)

    def _do_export_project(self, *, project_full_path: str, export_dst_full_path: Path, timeout_sec: int = 60, **kwargs):
        raise NotImplementedError()

    def export_project(self, *, project_full_path: str, export_dst_full_path: Path, timeout_sec: int = 60, **kwargs):
        self._logger.debug(f'Exporting project "{project_full_path=}" ({export_dst_full_path=}; {timeout_sec=}; {kwargs=}).')
        return self._do_export_project(project_full_path=project_full_path,
                                       export_dst_full_path=export_dst_full_path,
                                       timeout_sec=timeout_sec, **kwargs)

    def _do_import_project(self, *, parent_group_full_path: str,
                           new_project_name: str, new_project_path: str,
                           import_src_full_path: Path, timeout_sec: int = 60, **kwargs):
        raise NotImplementedError()

    def import_project(self, *, parent_group_full_path: str,
                       new_project_name: str, new_project_path: str,
                       import_src_full_path: Path, timeout_sec: int = 60, **kwargs):
        self._logger.debug(f'Importing project "{new_project_path=}" ({parent_group_full_path=}; {new_project_name=}; {import_src_full_path=}; {timeout_sec=}; {kwargs=}).')
        return self._do_import_project(parent_group_full_path=parent_group_full_path,
                                       new_project_name=new_project_name, new_project_path=new_project_path,
                                       import_src_full_path=import_src_full_path,
                                       timeout_sec=timeout_sec, **kwargs)
